/* LCGI - Linux C Graphics Interface */
/* Version: 1.0 (5.11.2001)          */
/* Autor: Dr. Helmut Herold          */
  
#include "graphics.h"

#include <stdarg.h>
#include <stdlib.h>
#include <qapplication.h>
#include <qpainter.h>
#include <qnamespace.h>
#include <qmessagebox.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qstring.h>
#include <qcursor.h>
#include <qinputdialog.h>
#if !defined(QT_H)
#include "qglobal.h"
#endif

#define MAX_PIXMAPS     10000

typedef enum { Arc,          Bar,           Bar3d,      Circle,
               Cleardevice,  Clearviewport, Drawpoly,   Ellipse,
               Fillellipse,  Fillpoly,      Line,       Linerel,
               Lineto,       Moverel,       Moveto,     Outtextxy,
               Pieslice,     Putimage,      Putpixel,   Rectangle,
               Setwritemode, Sector,        Setviewport           } typ;
  
//********************************************************************* Tabellen

QColor colorMap[] = {   //................................... colorMap
            QColor(   0,   0,   0 ),
            QColor(   0,   0, 255 ),
            QColor(   0, 255,   0 ),
            QColor(   0, 255, 255 ),
            QColor( 255,   0,   0 ),
            QColor( 255,   0, 255 ),
            QColor( 120,  58,  58 ),
            QColor( 211, 211, 211 ),
            QColor( 150, 150, 150 ),
            QColor( 155, 200, 255 ),
            QColor( 150, 255, 150 ),
            QColor( 200, 255, 255 ),
            QColor( 255,  80,  80 ),
            QColor( 255, 190, 255 ),
            QColor( 255, 255,   0 ),
            QColor( 255, 255, 255 ) };

Qt::BrushStyle brushMap[] = {   //........................... brushMap
      Qt::NoBrush,
      Qt::SolidPattern,
      Qt::Dense1Pattern,
      Qt::Dense2Pattern,
      Qt::Dense3Pattern,
      Qt::Dense4Pattern,
      Qt::Dense5Pattern,
      Qt::Dense6Pattern,
      Qt::Dense7Pattern,
      Qt::HorPattern,
      Qt::VerPattern,
      Qt::CrossPattern,
      Qt::BDiagPattern,
      Qt::FDiagPattern,
      Qt::DiagCrossPattern,
      Qt::CustomPattern     };

Qt::PenStyle lineMap[] = {   //............................... lineMap
      Qt::NoPen,
      Qt::SolidLine,
      Qt::DashLine,
      Qt::DotLine,
      Qt::DashDotLine,
      Qt::DashDotDotLine
};

Qt::AlignmentFlags hAlignMap[] = { //....................... hAlignMap
       Qt::AlignLeft,
       Qt::AlignHCenter,
       Qt::AlignRight
};

Qt::AlignmentFlags vAlignMap[] = { //....................... vAlignMap
       Qt::AlignBottom,
       Qt::AlignVCenter,
       Qt::AlignTop
};

Qt::RasterOp rasterMap[] = {  //............................ rasterMap
        Qt::CopyROP,
        Qt::XorROP,
        Qt::OrROP,
        Qt::AndROP,
        Qt::NotROP
};

int cursorMap[] = {  //..................................... cursorMap
    ArrowCursor,      // standard arrow cursor
    UpArrowCursor,    // upwards arrow
    CrossCursor,      // crosshair
    WaitCursor,       // hourglass/watch
    IbeamCursor,      // ibeam/text entry
    SizeVerCursor,    // vertical resize
    SizeHorCursor,    // horizontal resize
    SizeBDiagCursor,  // diagonal resize (/)
    SizeFDiagCursor,  // diagonal resize (\)
    SizeAllCursor,    // all directions resize
    BlankCursor,      // blank/invisible cursor
    SplitVCursor,     // vertical bar with left-right
    SplitHCursor,     // horizontal bar with up-down
    PointingHandCursor  // pointing hand
};

Qt::Key keyMap[] = {
    Qt::Key_Escape,      // misc keys
    Qt::Key_Tab,
    Qt::Key_Backtab,
    Qt::Key_Backspace,
    Qt::Key_Return,
    Qt::Key_Enter,
    Qt::Key_Insert,
    Qt::Key_Delete,
    Qt::Key_Pause,
    Qt::Key_Print,
    Qt::Key_SysReq,
    Qt::Key_Home,        // cursor movement
    Qt::Key_End,
    Qt::Key_Left,
    Qt::Key_Up,
    Qt::Key_Right,
    Qt::Key_Down,
    Qt::Key_PageUp,
    Qt::Key_PageDown,
    Qt::Key_Shift,        // modifiers
    Qt::Key_Control,
    Qt::Key_Meta,
    Qt::Key_Alt,
    Qt::Key_CapsLock,
    Qt::Key_NumLock,
    Qt::Key_ScrollLock,
    Qt::Key_F1,   // function keys
    Qt::Key_F2,
    Qt::Key_F3,
    Qt::Key_F4,
    Qt::Key_F5,
    Qt::Key_F6,
    Qt::Key_F7,
    Qt::Key_F8,
    Qt::Key_F9,
    Qt::Key_F10,
    Qt::Key_F11,
    Qt::Key_F12,
    Qt::Key_F13,
    Qt::Key_F14,
    Qt::Key_F15,
    Qt::Key_F16,
    Qt::Key_F17,
    Qt::Key_F18,
    Qt::Key_F19,
    Qt::Key_F20,
    Qt::Key_F21,
    Qt::Key_F22,
    Qt::Key_F23,
    Qt::Key_F24,
    Qt::Key_F25,   // F25 .. F35 only on X11
    Qt::Key_F26,
    Qt::Key_F27,
    Qt::Key_F28,
    Qt::Key_F29,
    Qt::Key_F30,
    Qt::Key_F31,
    Qt::Key_F32,
    Qt::Key_F33,
    Qt::Key_F34,
    Qt::Key_F35,
    Qt::Key_Super_L,    // extra keys
    Qt::Key_Super_R,
    Qt::Key_Menu,
    Qt::Key_Hyper_L,
    Qt::Key_Hyper_R,
    Qt::Key_Space,     // 7 bit printable ASCII
    Qt::Key_Exclam,
    Qt::Key_QuoteDbl,
    Qt::Key_NumberSign,
    Qt::Key_Dollar,
    Qt::Key_Percent,
    Qt::Key_Ampersand,
    Qt::Key_Apostrophe,
    Qt::Key_ParenLeft,
    Qt::Key_ParenRight,
    Qt::Key_Asterisk,
    Qt::Key_Plus,
    Qt::Key_Comma,
    Qt::Key_Minus,
    Qt::Key_Period,
    Qt::Key_Slash,
    Qt::Key_0,
    Qt::Key_1,
    Qt::Key_2,
    Qt::Key_3,
    Qt::Key_4,
    Qt::Key_5,
    Qt::Key_6,
    Qt::Key_7,
    Qt::Key_8,
    Qt::Key_9,
    Qt::Key_Colon,
    Qt::Key_Semicolon,
    Qt::Key_Less,
    Qt::Key_Equal,
    Qt::Key_Greater,
    Qt::Key_Question,
    Qt::Key_At,
    Qt::Key_A,
    Qt::Key_B,
    Qt::Key_C,
    Qt::Key_D,
    Qt::Key_E,
    Qt::Key_F,
    Qt::Key_G,
    Qt::Key_H,
    Qt::Key_I,
    Qt::Key_J,
    Qt::Key_K,
    Qt::Key_L,
    Qt::Key_M,
    Qt::Key_N,
    Qt::Key_O,
    Qt::Key_P,
    Qt::Key_Q,
    Qt::Key_R,
    Qt::Key_S,
    Qt::Key_T,
    Qt::Key_U,
    Qt::Key_V,
    Qt::Key_W,
    Qt::Key_X,
    Qt::Key_Y,
    Qt::Key_Z,
    Qt::Key_BracketLeft,
    Qt::Key_Backslash,
    Qt::Key_BracketRight,
    Qt::Key_AsciiCircum,
    Qt::Key_Underscore,
    Qt::Key_QuoteLeft,
    Qt::Key_BraceLeft,
    Qt::Key_Bar,
    Qt::Key_BraceRight,
    Qt::Key_AsciiTilde
};
const int keyMapSize = sizeof(keyMap) / sizeof(keyMap[0]);


//************************************************************ Klasse MalFenster
class MalFenster : public QWidget
{
   Q_OBJECT
public:
   MalFenster( int w=640, int h=480 );

     //............................................... Tastatur-Funktionen
   int    kbhit( void );
   int    getch( void );
   char   getcharacter( char *text );
   double getdouble( char *text );
   int    getint( char *text );
   char  *gettext( char *text );
     //......................... Bildschirm-, Farben- und Pixel-Funktionen
   void cleardevice( int color );
   int  getmaxcolor( void );
   int  getmaxx( void );
   int  getmaxy( void );
   int  getpixel( int x, int y );
   int  getx( void );
   int  gety( void );
   void putpixel( int x, int y, int color );
     //............... Positionieren, Linien zeichnen und Farbe einstellen
   int  getcolor(void);
   void moverel( int x1, int y1 );
   void moveto( int x1, int y1 );
   void line( int x1, int y1, int x2, int y2 );
   void linerel( int x1, int y1 );
   void lineto( int x1, int y1 );
   void setcolor( int color );
   void setlinestyle( int lstyle, int thickness );
     //................................................. Figuren-Funktionen
   void arc( int x, int y, int start, int end, int radius );
   void bar( int x1, int y1, int x2, int y2 );
   void bar3d( int x1, int y1, int x2, int y2, int depth );
   void circle( int x, int y, int radius );
   void drawpoly( int numpoints, int *polypoints );
   void ellipse( int x, int y, int start, int end, int xradius, int yradius );
   void fillellipse( int x, int y, int xradius, int yradius );
   void fillpoly( int numpoints, int *polypoints );
   void pieslice( int x, int y, int start, int end, int radius );
   void rectangle(int x1, int y1, int x2, int y2 );
   void sector( int x, int y, int start, int end, int xradius, int yradius );
   void setfillstyle( int pattern, int color );
     //................................................... Text-Funktionen
   void outtextxy( int x1, int y1, int x2, int y2, char *text );
   int  textheight( void );
   int  textwidth( char *text );
   void settextjustify( int hJust, int vJust );
   void settextstyle( char *name, int groesse );
     //................................................. Bilder-Funktionen
   void getimage( int left, int top, int right, int bottom, void **imagebuffer );
   void loadimage( char *filename, void **imagebuffer );
   void putimage( int x, int y, void *imagebuffer, int op );
   void freeimage( void **imagebuffer ); 
     //........................................... Einstellungs-Funktionen
   void setpalette( int farbe, int neufarbe );
   void setrgbpalette( int farbe, int r, int g, int b );
   void setwritemode( int op );
     //......................................... Zeichenfenster-Funktionen
   void clearviewport( int color );
   void setviewport(int x1, int y1, int x2, int y2 );
     //................................................... Maus-Funktionen
   void mouse_setpos( int x, int y );
   void mouse_getpos( int *x, int *y );
   int  mouse_left( void );
   int  mouse_mid( void );
   int  mouse_right( void );
   int  mouse_button( void );
   void mouse_show( void );
   void mouse_hide( void );
   int  mouse_visible( void );
   void mouse_setcursor( int mouseCursor );
   void mouse_setwindow( int x1, int y1, int x2, int y2 );
   int  mouse_inwindow( int x1, int y1, int x2, int y2 );

public slots:
   void readySlot( void );

protected:
   virtual void timerEvent( QTimerEvent *e );
   virtual void paintEvent( QPaintEvent *e );
   virtual void keyPressEvent( QKeyEvent *e );
   virtual void mousePressEvent( QMouseEvent *e );
   virtual void mouseReleaseEvent( QMouseEvent *e );
   virtual void mouseMoveEvent( QMouseEvent *e );
   virtual void showEvent( QShowEvent *e );
   virtual void moveEvent( QMoveEvent *e );
   virtual void resizeEvent( QResizeEvent *e );

private:
   void paint( void  );

   QPainter      *p;
   QPainter      *pp;
   QPixmap        puffer;
   QPoint         currentPos;
   int            width, height;
   int            x1, x2, y1, y2;
   int            radius;
   int            start;
   int            end;
   int            xradius;
   int            yradius;
   int            depth;
   QPointArray    points;
   int            noPoints;
   typ            type;
   int            keyPressed;
   int            key;
   QColor         color;
   int            colorId;
   QColor         bgColor;
   int            bgColorId;
   QColor         pixelColor;
   int            pixelColorId;
   QPointArray    pointArray;
   BrushStyle     fillStyle;
   QColor         fillColor;
   int            fillColorId;
   int            hJustId;
   int            vJustId;
   AlignmentFlags textAlign;
   QString        fontName;
   int            textGroesse;
   QString        currentText;
   Qt::PenStyle   lineStyle;
   int            lineThick;
   QPixmap        image;
   int            rasterOpId;
   Qt::RasterOp   rasterOp;
   int            opTemp;
   double         PI;
   int            clipx1, clipy1, clipx2, clipy2;
   int            mouseCursor;
   int            mouseLeft, mouseMid, mouseRight, mouseButton;
   int            mx1, my1, mx2, my2;
   bool           mwindow;
   QPixmap        pixmap[MAX_PIXMAPS];
   void          *pixmapAdresse[MAX_PIXMAPS];
   int            pixmapNumber;
   char           statischText[1000];
};

#include "graphics.moc"

//*************************************** Funktionen initgraph_ und closegraph

MalFenster *__myM;
QApplication *app;

void initgraph_( int argc, char *argv[], int w, int h ) {
    app = new QApplication( argc, argv );
    __myM = new MalFenster( w, h );
    QObject::connect( qApp, SIGNAL( lastWindowClosed() ),
                      __myM, SLOT( readySlot() ) );
}

void closegraph( void ) { delete app; }


//....................................................... MalFenster-Konstruktor
MalFenster::MalFenster( int w, int h ) : QWidget( 0 ) {
   colorId      = BLACK;
   color        = colorMap[BLACK];
   bgColorId    = LIGHTGRAY;
   bgColor      = colorMap[LIGHTGRAY];
   fillStyle    = brushMap[SOLID_FILL];
   fillColorId  = WHITE;
   fillColor    = colorMap[fillColorId];
   pixelColorId = BLACK;
   pixelColor   = colorMap[BLACK];
   lineStyle    = (Qt::PenStyle)SOLID_LINE;
   lineThick    = 1;
   hJustId      = LEFT_TEXT;
   vJustId      = TOP_TEXT;
   textAlign    = (AlignmentFlags)(hAlignMap[hJustId] | vAlignMap[vJustId]);
   fontName     = "Helvetica";
   textGroesse  = 12;
   rasterOpId   = COPY_PUT;
   rasterOp     = rasterMap[rasterOpId];
   keyPressed   = 0;
   currentPos   = QPoint( 0, 0 );
   clipx1 = clipy1 = 0;
   clipx2 = w-1;
   clipy2 = h-1;
   mouseCursor = cursorMap[ArrowCursor];
   mouseLeft = mouseMid = mouseRight = mouseButton = 0;
   mwindow = false;
   pixmapNumber = 0;

   p = new QPainter;
   pp = new QPainter;
   PI = 4*atan(1);
   resize( w, h );
   setFixedSize( w, h );
   width  = w;
   height = h;
   show();
}


//********************************************************* Tastatur-Funktionen

//....................................................................... kbhit
int MalFenster::kbhit( void ) {
   qApp->processEvents();
   return keyPressed;
}

//....................................................................... getch
int MalFenster::getch( void ) {
   qApp->processEvents();
   while ( !keyPressed )
      qApp->processEvents();
   keyPressed = 0;
   return key;
}

//................................................................ getcharacter
char MalFenster::getcharacter( char *info ) {
   bool ok = false;
   QString z =QInputDialog::getText( "Eingabe eines Zeichens", info, "", &ok, this );
   while ( kbhit() )
      getch();
   qApp->processEvents();
   if ( !ok || z.isEmpty() )
      return ' ';
   return (char )*(z.left(1).latin1());
}

//................................................................... getdouble
double MalFenster::getdouble( char *info ) {
   double z =QInputDialog::getDouble( "Eingabe einer Double-Zahl", info );
   while (kbhit())
      getch();
   qApp->processEvents();
   return z;
}

//...................................................................... getint
int MalFenster::getint( char *info ) {
   int z =QInputDialog::getInteger( "Eingabe einer Integer-Zahl", info );
   while (kbhit())
      getch();
   qApp->processEvents();
   return z;
}

//..................................................................... gettext
char *MalFenster::gettext( char *info ) {
   QString z = QInputDialog::getText( "Eingabe von Text", info );
   while (kbhit())
      getch();
   qApp->processEvents();
   for (unsigned i=0; i<z.length(); i++)
      statischText[i] = (char) *(z.mid(i,1).latin1());
   statischText[z.length()] = '\0';
   return (char *)statischText;
}


//*********************************** Bildschirm-, Farben- und Pixel-Funktionen

//................................................................. cleardevice
void MalFenster::cleardevice( int color ) {
   bgColorId = color;
   bgColor   = colorMap[color];
   type = Cleardevice;
   paint();
}

//................................................................. getmaxcolor
int MalFenster::getmaxcolor( void ) {
   return MAXCOLORS;
}

//..................................................................... getmaxx
int MalFenster::getmaxx( void ) {
   return width-1;
}

//..................................................................... getmaxy
int MalFenster::getmaxy( void ) {
   return height-1;
}

//.................................................................... getpixel
int MalFenster::getpixel( int x, int y ) {
   QPixmap puffer2;
   puffer2.resize(1, 1);
   bitBlt( &puffer2, 0, 0, &puffer, x, y, 1, 1, CopyROP );
   QColor c = QColor( puffer2.convertToImage().pixel(0,0) );
   
   for ( int i=0; i<=getmaxcolor(); i++)
      if ( c == colorMap[i] )
         return  i;
   return WHITE;
}

//........................................................................ getx
int MalFenster::getx( void ) {
   return p->pos().x();
}

//........................................................................ gety
int MalFenster::gety( void ) {
   return p->pos().y();
}

//.................................................................... putpixel
void MalFenster::putpixel( int x, int y, int color ) {
   this->x1 = x;
   this->y1 = y;
   this->pixelColorId = color;
   this->pixelColor = colorMap[color];
   type = Putpixel;
   paint();
}


//************************* Positionieren, Linien zeichnen und Farbe einstellen

//.................................................................... getcolor
int MalFenster::getcolor(void) {
   return colorId;
}

//........................................................................ line
void MalFenster::line( int x1, int y1, int x2, int y2 ) {
   this->x1 = x1;
   this->y1 = y1;
   this->x2 = x2;
   this->y2 = y2;
   type = Line;
   paint();
}

//..................................................................... linerel
void MalFenster::linerel( int x1, int y1 ) {
   this->x1 = x1;
   this->y1 = y1;
   type = Linerel;
   paint();
}

//...................................................................... lineto
void MalFenster::lineto( int x1, int y1 ) {
   this->x1 = x1;
   this->y1 = y1;
   type = Lineto;
   paint();
}

//..................................................................... moverel
void MalFenster::moverel( int x1, int y1 ) {
   this->x1 = x1;
   this->y1 = y1;
   type = Moverel;
   paint();
}

//...................................................................... moveto
void MalFenster::moveto( int x1, int y1 ) {
   this->x1 = x1;
   this->y1 = y1;
   type = Moveto;
   paint();
}

//.................................................................... setcolor
void MalFenster::setcolor( int color ) {
   this->colorId = color;
   this->color   = colorMap[color];
}

//................................................................ setlinestyle
void MalFenster::setlinestyle( int lstyle, int thickness ) {
   lineStyle = lineMap[lstyle];
   lineThick = thickness;
}



//********************************************************** Figuren-Funktionen

//......................................................................... arc
void MalFenster::arc( int x, int y, int start, int end, int radius ) {
   this->x1      = x;
   this->y1      = y;
   this->start   = start;
   this->end     = end;
   this->radius  = radius;
   type    =  Arc;
   paint();
}

//......................................................................... bar
void MalFenster::bar( int x1, int y1, int x2, int y2 ) {
   this->x1 = x1;
   this->y1 = y1;
   this->x2 = x2;
   this->y2 = y2;
   this->type = Bar;
   this->paint();
}

//....................................................................... bar3d
void MalFenster::bar3d( int x1, int y1, int x2, int y2, int depth ) {
   if ( x1 <= x2 )  {
       this->x1 = x1;
       this->x2 = x2;
   } else {
       this->x1 = x2;
       this->x2 = x1;
   }
   if ( y1 <= y2 )  {
       this->y1 = y1;
       this->y2 = y2;
   } else {
       this->y1 = y2;
       this->y2 = y1;
   }
   this->depth = depth;
   type = Bar3d;
   paint();
}

//...................................................................... circle
void MalFenster::circle( int x, int y, int radius ) {
   this->x1     = x;
   this->y1     = y;
   this->radius = radius;
   type =  Circle;
   paint();
}

//.................................................................... drawpoly
void MalFenster::drawpoly( int numpoints, int *polypoints ) {
   pointArray.resize( numpoints );
   for ( int i=0; i<numpoints*2; i+=2 ) {
      pointArray[ i/2] = QPoint( polypoints[i], polypoints[i+1] );
   }
   type = Drawpoly;
   paint();
}

//..................................................................... ellipse
void MalFenster::ellipse( int x, int y, int start, int end,
                          int xradius, int yradius ) {
   this->x1      = x;
   this->y1      = y;
   this->start   = start;
   this->end     = end;
   this->xradius = xradius;
   this->yradius = yradius;
   type =  Ellipse;
   paint();
}

//................................................................. fillellipse
void MalFenster::fillellipse( int x, int y, int xradius, int yradius ) {
   this->x1      = x;
   this->y1      = y;
   this->xradius = xradius;
   this->yradius = yradius;
   type =  Fillellipse;
   paint();
}

//.................................................................... fillpoly
void MalFenster::fillpoly( int numpoints, int *polypoints ) {
   pointArray.resize( numpoints );
   for ( int i=0; i<numpoints*2; i+=2 ) {
       pointArray[ i/2] = QPoint( polypoints[i], polypoints[i+1] );
   }
   type = Fillpoly;
   paint();
}

//.................................................................... pieslice
void MalFenster::pieslice( int x, int y, int start, int end, int radius ) {
   this->x1      = x;
   this->y1      = y;
   this->start   = start;
   this->end     = end;
   this->radius  = radius;
   type =  Pieslice;
   paint();
}

//................................................................... rectangle
void MalFenster::rectangle(int x1, int y1, int x2, int y2 ) {
   this->x1 = x1;
   this->y1 = y1;
   this->x2 = x2;
   this->y2 = y2;
   type = Rectangle;
   paint();
}

//...................................................................... sector
void MalFenster::sector( int x, int y, int start, int end,
                          int xradius, int yradius ) {
   this->x1      = x;
   this->y1      = y;
   this->start   = start;
   this->end     = end;
   this->xradius = xradius;
   this->yradius = yradius;
   type =  Sector;
   paint();
}

//................................................................ setfillstyle
void MalFenster::setfillstyle( int pattern, int color ) {
   fillStyle   = brushMap[pattern];
   fillColorId = color;
   fillColor   = colorMap[fillColorId];
}



//************************************************************* Text-Funktionen

//................................................................... outtextxy
void MalFenster::outtextxy( int x1, int y1, int x2, int y2, char *text ) {
   currentText = text;
   this->x1 = x1;
   this->y1 = y1;
   this->x2 = x2;
   this->y2 = y2;
   type = Outtextxy;
   paint();
}

//.................................................................. textheight
int MalFenster::textheight( void ) {
   int g;
   p->begin( this );
   p->setFont( QFont( fontName, textGroesse ) );
   g = p->fontMetrics().height();
   p->end();
   return g;
}

//................................................................... textwidth
int MalFenster::textwidth( char *text ) {
   int g;
   p->begin( this );
   p->setFont( QFont( fontName, textGroesse ) );
   g = p->fontMetrics().width( text );
   p->end();
   return g;
}

//.............................................................. settextjustify
void MalFenster::settextjustify( int hJust, int vJust ) {
   this->hJustId   = hJust;
   this->vJustId   = vJust;
   this->textAlign = (AlignmentFlags)(hAlignMap[hJustId] | vAlignMap[vJustId]);
}

//................................................................ settextstyle
void MalFenster::settextstyle( char *name, int groesse ) {
   this->fontName = name;
   this->textGroesse = groesse;
}



//********************************************************** Bilder-Funktionen

//.................................................................... getimage
void MalFenster::getimage( int left, int top, int right, int bottom,
                           void **imagebuffer ) {
   if (pixmapNumber >= 0 && pixmapNumber < MAX_PIXMAPS) {
      int w = right-left+1,
          h = bottom-top+1;
      pixmap[pixmapNumber].resize( w, h );
      bitBlt( &pixmap[pixmapNumber], 0, 0, &puffer, left, top, w, h, CopyROP );
      pixmapAdresse[pixmapNumber] = *imagebuffer =  &pixmap[pixmapNumber];
      pixmapNumber++;
   }
}

//................................................................... loadimage
void MalFenster::loadimage( char *filename, void **imagebuffer ) {
   if (pixmapNumber >= 0 && pixmapNumber < MAX_PIXMAPS) {
      if ( ! pixmap[pixmapNumber].load( filename ) ) {
         QString fehlerMeld;
         fehlerMeld.sprintf("cannot open file `%s'", filename );
         QMessageBox::warning( 0, "Load error", fehlerMeld );
      }
      pixmapAdresse[pixmapNumber] = *imagebuffer =  &pixmap[pixmapNumber];
      pixmapNumber++;
   }
}

//.................................................................... putimage
void MalFenster::putimage( int x, int y, void *imagebuffer, int op ) {
   for (int i=0; i<pixmapNumber; i++)
      if (pixmapAdresse[i] != NULL && pixmapAdresse[i] == imagebuffer &&
          !pixmap[i].isNull()) {
         this->x1     = x;
         this->y1     = y;
         this->image.resize( pixmap[i].size() );
         this->image  = pixmap[i];
         this->opTemp = op;
         type         = Putimage;
         paint();
         break;
      }
}

//................................................................... freeimage
void MalFenster::freeimage( void **imagebuffer ) {
   for (int i=0; i<pixmapNumber; i++)
      if (pixmapAdresse[i] != NULL && pixmapAdresse[i] == imagebuffer &&
          !pixmap[i].isNull()) {
         pixmap[pixmapNumber].resize( 0, 0 );         
         pixmapAdresse[pixmapNumber] = *imagebuffer = NULL;
         break;
      }
}


//***************************************************** Einstellungs-Funktionen

//.................................................................. setpalette
void MalFenster::setpalette( int farbe, int neufarbe ) {
   colorMap[farbe] = colorMap[neufarbe]; 
}

//............................................................... setrgbpalette
void MalFenster::setrgbpalette( int farbe, int r, int g, int b ) {
   colorMap[farbe] = QColor( r, g, b );
}

//................................................................ setwritemode
void MalFenster::setwritemode( int op ) {
   rasterOpId = op;
   rasterOp   = rasterMap[op];
}

//*************************************************** Zeichenfenster-Funktionen

//............................................................... clearviewport
void MalFenster::clearviewport( int color ) {
   bgColorId = color;
   bgColor   = colorMap[color];
   type = Clearviewport;
   paint();
}

//................................................................. setviewport
void MalFenster::setviewport(int x1, int y1, int x2, int y2 ) {
   this->clipx1 = x1;
   this->clipy1 = y1;
   this->clipx2 = x2;
   this->clipy2 = y2;
   type = Setviewport;
   paint();
}



//************************************************************* Maus-Funktionen

//................................................................ mouse_setpos
void MalFenster::mouse_setpos( int x, int y ) {
   QCursor::setPos( mapToGlobal( QPoint( x, y ) ) );
}
 
//................................................................ mouse_getpos
void MalFenster::mouse_getpos( int *x, int *y ) {
   QPoint p = mapFromGlobal( QCursor::pos() );
   *x = p.x();
   *y = p.y();
}

//.................................................................. mouse_left
int MalFenster::mouse_left( void ) {
   qApp->processEvents();
   return mouseLeft;
}

//................................................................... mouse_mid
int MalFenster::mouse_mid( void ) {
   qApp->processEvents();
   return mouseMid;
}

//................................................................. mouse_right
int MalFenster::mouse_right( void ) {
   qApp->processEvents();
   return mouseRight;
}

//................................................................ mouse_button
int MalFenster::mouse_button( void ) {
   qApp->processEvents();
   return mouseButton;
}
 
//.................................................................. mouse_show
void MalFenster::mouse_show( void ) {
   setCursor( cursorMap[mouseCursor] );
}

//.................................................................. mouse_hide
void MalFenster::mouse_hide( void ) {
   setCursor( cursorMap[BlankCursor] );
}

//............................................................... mouse_visible
int MalFenster::mouse_visible( void ) {
   if ( cursor().shape() == cursorMap[BlankCursor] )
      return 0;
   else
      return 1;
}

//............................................................. mouse_setcursor
void MalFenster::mouse_setcursor( int mouseCursor ) {
   setCursor( this->mouseCursor = mouseCursor );
}

//............................................................. mouse_setwindow
void MalFenster::mouse_setwindow( int x1, int y1, int x2, int y2 ) {
   mwindow = true;
   this->mx1 = x1;
   this->my1 = y1;
   this->mx2 = x2;
   this->my2 = y2;
}

//.............................................................. mouse_inwindow
int MalFenster::mouse_inwindow( int x1, int y1, int x2, int y2 ) {
   QPoint p = mapFromGlobal( QCursor::pos() );
   if ( p.x() >= x1 && p.y() >= y1 &&
        p.x() <= x2 && p.y() <= y2    )
      return 1;
   else
      return 0;
}



//******************************************************** protected-Funktionen

//................................................................... readySlot
void MalFenster::readySlot( void ) {
   exit( 1 );
}

//.................................................................. timerEvent
void MalFenster::timerEvent( QTimerEvent * ) {
   repaint( true );
   qApp->processEvents();
}

//.................................................................. paintEvent
void MalFenster::paintEvent( QPaintEvent *  ) {
   bitBlt( this, 0, 0, &puffer );
}

//............................................................... keyPressEvent
void MalFenster::keyPressEvent( QKeyEvent *e ) {
   keyPressed = 1;
   for ( int i=0; i<keyMapSize; i++)
      if ( keyMap[i] == e->key() ) {
         key = i;
         break;
      }
}

//............................................................. mousePressEvent
void MalFenster::mousePressEvent( QMouseEvent *e ) {
   if ( e->button() == LeftButton )
      mouseLeft = mouseButton = 1;
   else if ( e->button() == MidButton ) 
      mouseMid = mouseButton = 1;
   else if ( e->button() == RightButton ) 
      mouseRight = mouseButton = 1;
}

//........................................................... mouseReleaseEvent
void MalFenster::mouseReleaseEvent( QMouseEvent * ) {
   mouseLeft = mouseMid = mouseRight = mouseButton = 0;
}

//.............................................................. mouseMoveEvent
void MalFenster::mouseMoveEvent( QMouseEvent * ) {
   if ( mwindow ) {
      QPoint p = mapFromGlobal( QCursor::pos() );
      if ( p.x() < mx1 ) QCursor::setPos( mapToGlobal( QPoint( mx1, p.y() ) ) );
      if ( p.y() < my1 ) QCursor::setPos( mapToGlobal( QPoint( p.x(), my1 ) ) );
      if ( p.x() > mx2 ) QCursor::setPos( mapToGlobal( QPoint( mx2, p.y() ) ) );
      if ( p.y() > my2 ) QCursor::setPos( mapToGlobal( QPoint( p.x(), my2 ) ) );
   }
}

//................................................................... showEvent
void MalFenster::showEvent( QShowEvent * ) {
   bitBlt( this, 0, 0, &puffer );
   qApp->processEvents();
}

//................................................................... moveEvent
void MalFenster::moveEvent( QMoveEvent * ) {
   bitBlt( this, 0, 0, &puffer );
   qApp->processEvents();
}

//................................................................. resizeEvent
void MalFenster::resizeEvent( QResizeEvent *e ) {
   QPixmap save( puffer );
   puffer.resize( e->size() );
   puffer.fill( colorMap[bgColorId] );
   bitBlt( &puffer, 0, 0, &save );
   qApp->processEvents();
}

//....................................................................... paint
void MalFenster::paint( void  ) {
   int          x, y, len,
                top, left, right, bottom;
   QPointArray  poly;
   Qt::RasterOp opOld;

   p->begin( this );
   pp->begin( &puffer );
   p->setPen( QPen( colorMap[colorId], lineThick, (Qt::PenStyle)lineStyle ) );
   pp->setPen( QPen( colorMap[colorId], lineThick, (Qt::PenStyle)lineStyle ) );
   p->setBrush( QBrush( colorMap[fillColorId], (BrushStyle)fillStyle ) );
   pp->setBrush( QBrush( colorMap[fillColorId], (BrushStyle)fillStyle ) );
   QPen p2 = p->pen();
   p->setRasterOp( rasterOp );
   pp->setRasterOp( rasterOp );
   p->setClipRect( clipx1, clipy1, clipx2-clipx1+1, clipy2-clipy1+1 );
   pp->setClipRect( clipx1, clipy1, clipx2-clipx1+1, clipy2-clipy1+1 );

   switch ( type ) {

      case Arc:
         if ( (len = end-start) < 0 )
            len = 360 - (start - end);
         p->drawArc( x1-radius, y1-radius, radius*2+1, radius*2+1,
                     start*16, len*16 );
         pp->drawArc( x1-radius, y1-radius, radius*2+1, radius*2+1,
                      start*16, len*16 );
         break;

      case Bar:
         p->fillRect( x1, y1, x2-x1+1, y2-y1+1,
                      QBrush( colorMap[fillColorId], (BrushStyle)fillStyle ) );
         pp->fillRect( x1, y1, x2-x1+1, y2-y1+1,
                       QBrush( colorMap[fillColorId], (BrushStyle)fillStyle ) );
         break;

      case Bar3d:
         p->fillRect( x1, y1, x2-x1+1, y2-y1+1,
                      QBrush( colorMap[fillColorId], fillStyle ) );
         pp->fillRect( x1, y1, x2-x1+1, y2-y1+1,
                       QBrush( colorMap[fillColorId], fillStyle ) );
         p->drawRect( x1, y1, x2-x1+1, y2-y1+1 );
         pp->drawRect( x1, y1, x2-x1+1, y2-y1+1 );
         if ( depth > 0 ) {
            left = x1;
            top  = y1;
            right  = x2;
            bottom = y2;
            int dx = (int)(cos( PI/8 ) * depth);
            int dy = (int)(sin( PI/8 ) * depth);
            p->drawLine( left,  top, left+dx,  top-dy );
            pp->drawLine( left,  top, left+dx,  top-dy );
            p->drawLine( right, top, right+dx, top-dy );
            pp->drawLine( right, top, right+dx, top-dy );
            p->drawLine( left+dx, top-dy, right+dx, top-dy );
            pp->drawLine( left+dx, top-dy, right+dx, top-dy );
            p->drawLine( right, bottom, right+dx, bottom-dy );
            pp->drawLine( right, bottom, right+dx, bottom-dy );
            p->drawLine( right+dx, top-dy, right+dx, bottom-dy );
            pp->drawLine( right+dx, top-dy, right+dx, bottom-dy );
         }
         break;

      case Circle:
         p->drawArc( x1-radius, y1-radius, radius*2+1, radius*2+1, 0, 16*360 );
         pp->drawArc( x1-radius, y1-radius, radius*2+1, radius*2+1, 0, 16*360 );
         break;

      case Cleardevice:
         p->fillRect( 0, 0, width, height, colorMap[bgColorId] );
         pp->fillRect( 0, 0, width, height, colorMap[bgColorId] );
         break;

      case Clearviewport:
         p->fillRect( clipx1, clipy1, clipx2-clipx1+1, clipy2-clipy1+1,
                      colorMap[bgColorId] );
         pp->fillRect( clipx1, clipy1, clipx2-clipx1+1, clipy2-clipy1+1,
                       colorMap[bgColorId] );
         break;

      case Drawpoly:
         p->drawPolyline( pointArray );
         pp->drawPolyline( pointArray );
         break;

      case Ellipse:
         if ( (len = end-start) < 0 )
            len = 360 - (start - end);
         p->drawArc( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1,
                     start*16, len*16 );
         pp->drawArc( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1,
                      start*16, len*16 );
         break;

      case Fillellipse:
         p->drawEllipse( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1 );
         pp->drawEllipse( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1 );
         break;

      case Fillpoly:
         p->drawPolygon( pointArray );
         pp->drawPolygon( pointArray );
         break;

      case Line:
         p->drawLine( x1, y1, x2, y2 );
         pp->drawLine( x1, y1, x2, y2 );
         break;

      case Linerel:
         x = currentPos.x();
         y = currentPos.y();
         p->drawLine( x, y, x+x1, y+y1 );
         pp->drawLine( x, y, x+x1, y+y1 );
         break;

      case Lineto:
         x = currentPos.x();
         y = currentPos.y();
         p->drawLine( x, y, x1, y1 );
         pp->drawLine( x, y, x1, y1 );
         break;

      case Moverel:
         x = currentPos.x();
         y = currentPos.y();
         p->moveTo( x+x1, y+y1 );
         pp->moveTo( x+x1, y+y1 );
         break;

      case Moveto:
         p->moveTo( x1, y1 );
         pp->moveTo( x1, y1 );
         break;

      case Outtextxy:
         p->setPen( colorMap[colorId] );
         pp->setPen( colorMap[colorId] );
         p->setFont( QFont( fontName, textGroesse ) );
         pp->setFont( QFont( fontName, textGroesse ) );
         p->drawText( x1, y1, x2-x1+1, y2-y1+1,
                      textAlign|Qt::WordBreak, currentText );
         pp->drawText( x1, y1, x2-x1+1, y2-y1+1,
                       textAlign|Qt::WordBreak, currentText );
         break;

      case Pieslice:
         if ( (len = end-start) < 0 )
            len = 360 - (start - end);
         p->drawPie( x1-radius, y1-radius, radius*2+1, radius*2+1,
                     start*16, len*16 );
         pp->drawPie( x1-radius, y1-radius, radius*2+1, radius*2+1,
                      start*16, len*16 );
         break;

      case Putimage:
         opOld = p->rasterOp();
         p->setRasterOp( rasterMap[opTemp] );
         pp->setRasterOp( rasterMap[opTemp] );
         p->drawPixmap ( x1, y1, image );
         pp->drawPixmap ( x1, y1, image );
         p->setRasterOp( opOld );
         pp->setRasterOp( opOld );
         break;

      case Putpixel:
         p->setPen( colorMap[pixelColorId] );
         pp->setPen( colorMap[pixelColorId] );
         p->drawPoint( x1, y1 );
         pp->drawPoint( x1, y1 );
         break;

      case Rectangle:
         poly.resize( 5 );
         poly[0] = QPoint( x1, y1 );
         poly[1] = QPoint( x2, y1 );
         poly[2] = QPoint( x2, y2 );
         poly[3] = QPoint( x1, y2 );
         poly[4] = QPoint( x1, y1 );
         p->drawPolyline( poly );
         pp->drawPolyline( poly );
         break;

      case Sector:
         if ( (len = end-start) < 0 )
            len = 360 - (start - end);
         p->drawPie( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1,
                     start*16, len*16 );
         pp->drawPie( x1-xradius, y1-yradius, xradius*2+1, yradius*2+1,
                      start*16, len*16 );
         break;

      case Setviewport:
         p->setClipRect( clipx1, clipy1,
                         clipx2-clipx1, clipy2-clipy1 );
         pp->setClipRect( clipx1, clipy1,
                          clipx2-clipx1, clipy2-clipy1 );
         break;

      case Setwritemode:
         p->setRasterOp( rasterOp );
         pp->setRasterOp( rasterOp );
         break;
   }
   currentPos = p->pos();
   p->end();
   pp->end();
   qApp->processEvents();
}


//********************************************************* Tastatur-Funktionen

int  kbhit( void )  { return __myM->kbhit();     }
int  getch( void )  { return __myM->getch();     }

char getcharacter( char *text, ... ) {
   va_list     az;
   char puffer[5000];
   va_start(az, text);
   vsprintf(puffer, text, az);
   va_end(az);
   return __myM->getcharacter( puffer );
}

double getdouble( char *text, ... ) {
   va_list     az;
   char puffer[5000];
   va_start(az, text);
   vsprintf(puffer, text, az);
   va_end(az);
   return __myM->getdouble( puffer );
}

int getint( char *text, ... ) {
   va_list     az;
   char puffer[5000];
   va_start(az, text);
   vsprintf(puffer, text, az);
   va_end(az);
   return __myM->getint( puffer );
}

char *gettext( char *text, ... ) {
   va_list     az;
   char puffer[5000];
   va_start(az, text);
   vsprintf(puffer, text, az);
   va_end(az);
   return __myM->gettext( puffer );
}


//*********************************** Bildschirm-, Farben- und Pixel-Funktionen

void cleardevice( int color )   { __myM->cleardevice( color );    }
int  getmaxcolor( void )        { return __myM->getmaxcolor();    }
int  getmaxx( void )            { return __myM->getmaxx();        }
int  getmaxy( void )            { return __myM->getmaxy();        }
int  getpixel( int x, int y )   { return __myM->getpixel( x, y ); }
int  getx( void )               { return __myM->getx();           }
int  gety( void )               { return __myM->gety();           }

void putpixel( int x, int y, int color ) {
   __myM->putpixel( x, y, color );
}


//************************* Positionieren, Linien zeichnen und Farbe einstellen

int  getcolor(void)                         { return __myM->getcolor();      }
void line( int x1, int y1, int x2, int y2 ) { __myM->line( x1, y1, x2, y2 ); }
void linerel( int x1, int y1 )              { __myM->linerel( x1, y1 );      }
void lineto( int x1, int y1 )               { __myM->lineto( x1, y1 );       }
void moverel( int x1, int y1 )              { __myM-> moverel( x1, y1 );     }
void moveto( int x1, int y1 )               { __myM->moveto( x1, y1 );       }
void setcolor( int color )                  { __myM->setcolor( color );      }
void setlinestyle( int lstyle, int thick )  { __myM->setlinestyle( lstyle, thick ); }


//********************************************************** Figuren-Funktionen

void arc( int x, int y, int start, int end, int radius ) {
   __myM->arc( x, y, start, end, radius );
}

void bar( int x1, int y1, int x2, int y2 ) {
   __myM->bar( x1, y1, x2, y2 );
}

void bar3d( int x1, int y1, int x2, int y2, int depth ) {
   __myM->bar3d( x1, y1, x2, y2, depth );
}

void circle( int x, int y, int radius ) {
   __myM->circle( x, y, radius );
}

void drawpoly( int numpoints, int *polypoints ) {
   __myM->drawpoly( numpoints, polypoints );
}

void ellipse( int x, int y, int start, int end, int xradius, int yradius ) {
   __myM->ellipse( x, y, start, end, xradius, yradius );
}

void fillellipse( int x, int y, int xradius, int yradius ) {
   __myM->fillellipse( x, y, xradius, yradius );
}

void fillpoly( int numpoints, int *polypoints ) {
   __myM->fillpoly( numpoints, polypoints );
}

void pieslice( int x, int y, int start, int end, int radius ) {
   __myM->pieslice( x, y, start, end, radius );
}

void rectangle(int x1, int y1, int x2, int y2 ) {
   __myM->rectangle( x1, y1, x2, y2 );
}

void sector( int x, int y, int start, int end, int xradius, int yradius ) {
   __myM->sector( x, y, start, end, xradius, yradius );
}

void setfillstyle( int pattern, int color ) {
   __myM->setfillstyle( pattern, color );
}


//************************************************************* Text-Funktionen

void outtextxy( int x1, int y1, int x2, int y2, char *text, ... ) {
   va_list     az;
   char puffer[5000];
   va_start(az, text);
   vsprintf(puffer, text, az);
   __myM->outtextxy( x1, y1, x2, y2, puffer );
   va_end(az);
}

int  textheight( void )      { return __myM->textheight();      }
int  textwidth( char *text ) { return __myM->textwidth( text ); }

void settextjustify( int hJust, int vJust ) {
   __myM->settextjustify( hJust, vJust );
}

void settextstyle( char *name, int groesse ) {
   __myM->settextstyle( name, groesse );
}


//*********************************************************** Bilder-Funktionen

void getimage( int left, int top, int right, int bottom, void **imagebuffer ) {
   __myM->getimage( left, top, right, bottom, imagebuffer );
}

void loadimage( char *filename, void **imagebuffer ) {
   __myM->loadimage( filename, imagebuffer );
}

void putimage( int x, int y, void *imagebuffer, int op ) {
   __myM->putimage( x, y, imagebuffer, op );
}

void freeimage( void **imagebuffer ) {
   __myM->freeimage( imagebuffer );
}



//***************************************************** Einstellungs-Funktionen
 
void setpalette( int farbe, int neufarbe ) {
   __myM->setpalette( farbe, neufarbe );
}

void setrgbpalette( int farbe, int r, int g, int b ) {
   __myM->setrgbpalette( farbe, r, g, b );
}

void setwritemode( int op ) { __myM->setwritemode( op ); }
 


//*************************************************** Zeichenfenster-Funktionen

void clearviewport( int color ) { __myM->clearviewport( color ); }

void setviewport(int x1, int y1, int x2, int y2 ) {
   __myM->setviewport( x1, y1, x2, y2 );
}



//************************************************************* Maus-Funktionen
void mouse_setpos( int x, int y )   { __myM->mouse_setpos( x, y ); }
void mouse_getpos( int *x, int *y ) { __myM->mouse_getpos( x, y ); }

int  mouse_left( void )   { return __myM->mouse_left();   }
int  mouse_mid( void )    { return __myM->mouse_mid();    }
int  mouse_right( void )  { return __myM->mouse_right();  }
int  mouse_button( void ) { return __myM->mouse_button(); }

void mouse_show( void )    { __myM->mouse_show();           }
void mouse_hide( void )    { __myM->mouse_hide();           }
int  mouse_visible( void ) { return __myM->mouse_visible(); }

void mouse_setcursor( int mouseCursor ) {
   __myM->mouse_setcursor( mouseCursor );
}

void mouse_setwindow( int x1, int y1, int x2, int y2 ) {
   __myM->mouse_setwindow( x1, y1, x2, y2 );
}

int  mouse_inwindow( int x1, int y1, int x2, int y2 ) {
   return __myM->mouse_inwindow( x1, y1, x2, y2 );
}
